#! /usr/bin/env python

"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import sys
import traceback

from problog.program import PrologFile
from problog.constraint import TrueConstraint
from problog.formula import LogicFormula, LogicDAG
from problog.cnf_formula import CNF
from problog.maxsat import get_solver, get_available_solvers
from problog.errors import process_error


def main(argv, result_handler=None):
    args = argparser().parse_args(argv)
    inputfile = args.inputfile

    if result_handler is None:
        if args.web:
            result_handler = print_result_json
        else:
            result_handler = print_result

    if args.output is not None:
        outf = open(args.output, 'w')
    else:
        outf = sys.stdout

    try:
        pl = PrologFile(inputfile)

        dag = LogicDAG.createFrom(pl)

        cnf = CNF.createFrom(dag)

        for qn, qi in cnf.evidence():
            cnf.add_constraint(TrueConstraint(qi))

        for qn, qi in cnf.queries():
            cnf.add_constraint(TrueConstraint(qi))

        solver = get_solver(args.solver)

        result = frozenset(solver.evaluate(cnf))

        output_facts = None
        if result is not None:
            output_facts = []
            for i, n, t in dag:
                if t == 'atom':
                    if i in result:
                        output_facts.append(n.name)
                    elif args.output_all and -i in result:
                        output_facts.append(-n.name)

        result_handler((True, output_facts), outf)
    except Exception as err:
        trace = traceback.format_exc()
        err.trace = trace
        result_handler((False, err), outf)

    if args.output is not None:
        outf.close()


def print_result(result, output=sys.stdout):
    success, result = result
    if success:
        if result is None:
            print ('%% The model is not satisfiable.', file=output)
        else:
            for atom in result:
                print(atom, file=output)
        return 0
    else:
        print(process_error(result), file=output)
        return 1


def print_result_json(d, output):
    """Pretty print result.

    :param d: result from run_problog
    :param output: output file
    :param precision:
    :return:
    """
    import json
    result = {}
    success, d = d
    if success:
        result['SUCCESS'] = True
        if d is not None:
            result['atoms'] = list(map(lambda n: (str(-n), False) if n.is_negated() else (str(n), True), d))
    else:
        result['SUCCESS'] = False
        result['err'] = process_error(d)
        result['original'] = str(d)
    print (json.dumps(result), file=output)
    return 0


def reduce_formula(formula, facts):
    # Assume formula is cycle free.

    values = [None] * len(formula)
    for f in facts:
        if f > 0:
            values[f - 1] = True
        else:
            values[-f - 1] = False

    from collections import deque
    nodes = deque(range(1, len(formula) + 1))
    while nodes:
        index = nodes.popleft()
        value = values[index - 1]
        if value is None:
            node = formula.get_node(index)
            nodetype = type(node).__name__
            if nodetype == 'atom':
                pass  # Really shouldn't happen
            else:
                children = [values[c] if c > 0 else not values[-c] for c in node.children]
                if nodetype == 'disj':
                    if True in children:
                        values[index - 1] = True
                    elif None in children:
                        # not ready yet, push it back on the queue
                        nodes.append(index)
                    else:
                        values[index - 1] = False
                else:
                    if False in children:
                        values[index - 1] = False
                    elif None in children:
                        nodes.append(index)
                    else:
                        values[index - 1] = True
    return values


def argparser():
    import argparse

    parser = argparse.ArgumentParser()
    parser.add_argument('inputfile')
    parser.add_argument('--solver', choices=get_available_solvers(),
                        default=None, help="MaxSAT solver to use")
    parser.add_argument('--full', dest='output_all', action='store_true',
                        help='Also show false atoms.')
    parser.add_argument('-o', '--output', type=str, default=None,
                        help='Write output to given file (default: write to stdout)')
    parser.add_argument('--web', action='store_true', help=argparse.SUPPRESS)
    return parser


if __name__ == '__main__':
    main(sys.argv[1:])



"""
ProbLog command-line interface.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from __future__ import print_function

import os
import sys

from problog.formula import LogicDAG, LogicFormula
from problog.evaluator import SemiringLogProbability
from problog.parser import DefaultPrologParser
from problog.program import ExtendedPrologFactory, PrologFile
from problog.cnf_formula import CNF
from problog.errors import process_error
from problog.util import subprocess_check_output, mktempfile


def main(argv, result_handler=None):
    import argparse
    parser = argparse.ArgumentParser()
    parser.add_argument('filename', metavar='MODEL', type=str, help='input ProbLog model')
    parser.add_argument('--format', choices=('dot', 'pl', 'cnf', 'svg', 'internal'), default=None,
                        help='output format')
    parser.add_argument('--break-cycles', action='store_true', help='perform cycle breaking')
    parser.add_argument('--keep-all', action='store_true', help='also output deterministic nodes')
    parser.add_argument('--keep-duplicates', action='store_true', help='don\'t eliminate duplicate literals')
    parser.add_argument('--any-order', action='store_true', help='allow reordering nodes')
    parser.add_argument('--hide-builtins', action='store_true', help='hide deterministic part based on builtins')
    parser.add_argument('--propagate-evidence', action='store_true', help='propagate evidence')
    parser.add_argument('--propagate-weights', action='store_true', help='propagate evidence')
    parser.add_argument('--compact', action='store_true',
                        help='allow compact model (may remove some predicates)')
    parser.add_argument('--noninterpretable', action='store_true')
    parser.add_argument('--web', action='store_true', help=argparse.SUPPRESS)
    parser.add_argument('-o', '--output', type=str, help='output file', default=None)
    args = parser.parse_args(argv)

    outformat = args.format
    outfile = sys.stdout
    if args.output:
        outfile = open(args.output, 'w')
        if outformat is None:
            outformat = os.path.splitext(args.output)[1][1:]

    if outformat == 'cnf' and not args.break_cycles:
        print('Warning: CNF output requires cycle-breaking; cycle breaking enabled.',
              file=sys.stderr)

    if args.break_cycles or outformat == 'cnf':
        target = LogicDAG
    else:
        target = LogicFormula

    if args.propagate_weights:
        semiring = SemiringLogProbability()
    else:
        semiring = None

    if args.web:
        print_result = print_result_json
    else:
        print_result = print_result_standard

    try:
        gp = target.createFrom(
            PrologFile(args.filename, parser=DefaultPrologParser(ExtendedPrologFactory())),
            label_all=not args.noninterpretable, avoid_name_clash=not args.compact,
            keep_order=not args.any_order,
            keep_all=args.keep_all, keep_duplicates=args.keep_duplicates,
            hide_builtins=args.hide_builtins,
            propagate_evidence=args.propagate_evidence, propagate_weights=semiring)

        if outformat == 'pl':
            rc = print_result((True, gp.to_prolog()), output=outfile)
        elif outformat == 'dot':
            rc = print_result((True, gp.to_dot()), output=outfile)
        elif outformat == 'svg':
            dot = gp.to_dot()
            tmpfile = mktempfile('.dot')
            with open(tmpfile, 'w') as f:
                print(dot, file=f)
            svg = subprocess_check_output(['dot', tmpfile, '-Tsvg'])
            rc = print_result((True, svg), output=outfile)
        elif outformat == 'cnf':
            rc = print_result((True, CNF.createFrom(gp).to_dimacs()), output=outfile)
        elif outformat == 'internal':
            rc = print_result((True, str(gp)), output=outfile)
        else:
            rc = print_result((True, gp.to_prolog()), output=outfile)
    except Exception as err:
        import traceback
        err.trace = traceback.format_exc()
        rc = print_result((False, err))

    if args.output:
        outfile.close()

    if rc:
        sys.exit(rc)


def print_result_standard(result, output=sys.stdout):
    success, result = result
    if success:
        print (result, file=output)
        return 0
    else:
        print (process_error(result), file=output)
        return 1


def print_result_json(result, output=sys.stdout):
    success, result = result
    import json
    out = {'SUCCESS': success}
    if success:
        out['result'] = str(result)
    else:
        out['err'] = vars(result)
    print (json.dumps(out), file=output)
    return 0



"""
Module name
"""

from __future__ import print_function

from problog import root_path
from problog.util import subprocess_call, subprocess_check_output
import unittest
import os
import sys


class TestInterfaces(unittest.TestCase):

    def _test_cmd(self, task, testfile=None):
        problogcli = root_path('problog-cli.py')
        if testfile is None:
            testfile = [root_path('test', '7_probabilistic_graph.pl')]

        with open(os.devnull, 'w') as out:
            if task is None:
                self.assertEqual(
                    subprocess_call([sys.executable, problogcli] + testfile, stdout=out), 0)
            else:
                self.assertEqual(
                    subprocess_call([sys.executable, problogcli, task] + testfile, stdout=out), 0)

    def test_cli_mpe(self):
        return self._test_cmd('mpe')

    def test_cli_prob(self):
        return self._test_cmd('prob')

    def test_cli_explain(self):
        return self._test_cmd('explain')

    def test_cli_ground(self):
        return self._test_cmd('ground')

    def test_cli_sample(self):
        return self._test_cmd('sample')

    def test_cli_default(self):
        return self._test_cmd(None)

    def test_cli_learn(self):
        problogcli = root_path('problog-cli.py')

        model = root_path('problog', 'learning', 'test1_model.pl')
        examples = root_path('problog', 'learning', 'test1_examples.pl')

        out = subprocess_check_output([sys.executable, problogcli, 'lfi', model, examples])
        outline = out.strip().split()

        self.assertGreater(int(outline[-1]), 2)

        weights = [float(outline[i].strip('[],')) for i in (1, 2, 3, 4)]

        self.assertAlmostEqual(weights[0], 1.0 / 3)
        self.assertAlmostEqual(weights[2], 1.0)
        self.assertAlmostEqual(weights[3], 0.0)



"""
Part of the ProbLog distribution.

Copyright 2015 KU Leuven, DTAI Research Group

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""
import os
import sys

sys.setrecursionlimit(10000)

# Set the PATH and PYTHON_PATH variables
from .setup import set_environment, gather_info

set_environment()

system_info = gather_info()


def root_path(*args):
    return os.path.abspath(os.path.join(os.path.dirname(__file__), '..', *args))


# Load all submodules. This has two reasons:
#   - initializes all transformations (@transform)
#   - makes it possible to just import 'problog' and then use
#       something like problog.program.PrologFile
from . import cnf_formula
from . import core
from . import engine
from . import evaluator
from . import formula
from . import logic
from . import nnf_formula
from . import parser
from . import program
from . import sdd_formula
from . import util
from . import bdd_formula
from . import forward
from . import cycles
from . import kbest
from . import tasks


_evaluatables = {'sdd': sdd_formula.SDD,
                 'bdd': bdd_formula.BDD,
                 'nnf': nnf_formula.NNF,
                 'ddnnf': nnf_formula.NNF,
                 'kbest': kbest.KBestFormula,
                 'fsdd': forward.ForwardSDD,
                 'fbdd': forward.ForwardBDD}


def get_evaluatables():
    return _evaluatables.keys()


def get_evaluatable(name=None):
    if name is None:
        return evaluator.Evaluatable
    else:
        return _evaluatables[name]

